package transcoder

import (
	"bytes"
	"flag"
	"fmt"
	"io/ioutil"
	"sync"
	"text/template"

	pb "git.autistici.org/ale/autoradio/proto"
)

var (
	liquidsoapBin          = flag.String("liquidsoap", "/usr/bin/liquidsoap", "Location of the liquidsoap binary")
	liquidsoapTemplateFile = flag.String("liquidsoap-template", "", "Custom liquidsoap config template (if empty, use builtin default)")

	liquidsoapConfigStr = `
set("log.file", false)
set("log.stdout", true)
set("log.unix_timestamps", false)

upstream = mksafe(input.http("{{.SourceURL}}", buffer=5.0))
{{if eq .Channels 1}}upstream = mean(upstream){{end}}
output.icecast(%{{.Format}}(samplerate={{.SampleRate}}, {{if gt .BitRate 0}}bitrate={{.BitRate}}, {{end}}{{if gt .Quality -1.0}}quality={{.Quality}}, {{end}}{{if .StereoMode}}stereo_mode="{{.StereoMode}}", {{end}}{{if eq .Channels 2}}stereo{{else}}mono{{end}}),
  mount="{{.TargetMount}}", host="{{.TargetIP}}", port={{.TargetPort}}, user="{{.TargetUsername}}", password="{{.TargetPassword}}",
  upstream)
`

	liquidsoapConfigTpl *template.Template
	liquidsoapTplInit   sync.Once
)

func getLiquidsoapConfigTemplate() *template.Template {
	liquidsoapTplInit.Do(func() {
		tplStr := liquidsoapConfigStr
		if *liquidsoapTemplateFile != "" {
			data, err := ioutil.ReadFile(*liquidsoapTemplateFile)
			if err != nil {
				panic(err)
			}
			tplStr = string(data)
		}
		liquidsoapConfigTpl = template.Must(template.New("liquidsoap").Parse(tplStr))
	})
	return liquidsoapConfigTpl
}

// Parameters to configure a liquidsoap-based transcoder.
type liquidsoapParams struct {
	// Source (upstream) URL.
	SourceURL string

	// Target (downstream) connection parameters.
	TargetIP       string
	TargetPort     int
	TargetMount    string
	TargetUsername string
	TargetPassword string

	// Stream encoding parameters. One note on the 'float32'
	// choice for Quality: text/template can't evaluate the 'gt'
	// function successfully on a float64 type!
	Format     string
	BitRate    int32
	SampleRate int32
	Channels   int32
	StereoMode string
	Quality    float32
}

// Create new parameters for liquidsoap for a transcoding mount. If
// mount.TranscodeParams is nil, this function will panic, so the
// caller should check mount.Transcode.
//
// The source URL points at localhost, relying on the front-end
// proxying to actually send the traffic to the master. This has the
// advantage that we don't have to reconfigure liquidsoap on
// mastership changes.
func newLiquidsoapParams(m *pb.Mount) *liquidsoapParams {
	return &liquidsoapParams{
		SourceURL:      fmt.Sprintf("http://localhost%s", m.TranscodeParams.SourcePath),
		TargetIP:       "localhost",
		TargetPort:     80,
		TargetMount:    m.Path,
		TargetUsername: m.SourceUsername,
		TargetPassword: m.SourcePassword,
		Format:         m.TranscodeParams.Format,
		BitRate:        m.TranscodeParams.BitRate,
		SampleRate:     m.TranscodeParams.SampleRate,
		Channels:       m.TranscodeParams.Channels,
		StereoMode:     m.TranscodeParams.StereoMode,
		Quality:        m.TranscodeParams.Quality,
	}
}

func makeLiquidsoapConfig(m *pb.Mount) ([]byte, error) {
	var buf bytes.Buffer
	if err := getLiquidsoapConfigTemplate().Execute(&buf, newLiquidsoapParams(m)); err != nil {
		return nil, err
	}
	return buf.Bytes(), nil
}
